import logging

from sqlalchemy import (
    and_,
    false,
    or_,
)

import tool_shed.repository_types.util as rt_util
from tool_shed.util import (
    hg_util,
    metadata_util,
)
from tool_shed.webapp import model

log = logging.getLogger(__name__)


class Registry:
    def __init__(self, app):
        log.debug("Loading the repository registry...")
        self.app = app
        self.certified_level_one_clause_list = self.get_certified_level_one_clause_list()
        # The following lists contain tuples like ( repository.name, repository.user.username, changeset_revision )
        # where the changeset_revision entry is always the latest installable changeset_revision..
        self.certified_level_one_repository_and_suite_tuples = []
        self.certified_level_one_suite_tuples = []
        # These category dictionaries contain entries where the key is the category and the value is the integer count
        # of viewable repositories within that category.
        self.certified_level_one_viewable_repositories_and_suites_by_category = {}
        self.certified_level_one_viewable_suites_by_category = {}
        self.certified_level_two_repository_and_suite_tuples = []
        self.certified_level_two_suite_tuples = []
        self.certified_level_two_viewable_repositories_and_suites_by_category = {}
        self.certified_level_two_viewable_suites_by_category = {}
        self.repository_and_suite_tuples = []
        self.suite_tuples = []
        self.viewable_repositories_and_suites_by_category = {}
        self.viewable_suites_by_category = {}
        self.viewable_valid_repositories_and_suites_by_category = {}
        self.viewable_valid_suites_by_category = {}
        self.load_viewable_repositories_and_suites_by_category()
        self.load_repository_and_suite_tuples()

    def add_category_entry(self, category):
        category_name = str(category.name)
        if category_name not in self.viewable_repositories_and_suites_by_category:
            self.viewable_repositories_and_suites_by_category[category_name] = 0
        if category_name not in self.viewable_suites_by_category:
            self.viewable_suites_by_category[category_name] = 0
        if category_name not in self.viewable_valid_repositories_and_suites_by_category:
            self.viewable_valid_repositories_and_suites_by_category[category_name] = 0
        if category_name not in self.viewable_valid_suites_by_category:
            self.viewable_valid_suites_by_category[category_name] = 0
        if category_name not in self.certified_level_one_viewable_repositories_and_suites_by_category:
            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 0
        if category_name not in self.certified_level_one_viewable_suites_by_category:
            self.certified_level_one_viewable_suites_by_category[category_name] = 0

    def add_entry(self, repository):
        try:
            if repository:
                is_valid = self.is_valid(repository)
                certified_level_one_tuple = self.get_certified_level_one_tuple(repository)
                latest_installable_changeset_revision, is_level_one_certified = certified_level_one_tuple
                for rca in repository.categories:
                    category = rca.category
                    category_name = str(category.name)
                    if category_name in self.viewable_repositories_and_suites_by_category:
                        self.viewable_repositories_and_suites_by_category[category_name] += 1
                    else:
                        self.viewable_repositories_and_suites_by_category[category_name] = 1
                    if is_valid:
                        if category_name in self.viewable_valid_repositories_and_suites_by_category:
                            self.viewable_valid_repositories_and_suites_by_category[category_name] += 1
                        else:
                            self.viewable_valid_repositories_and_suites_by_category[category_name] = 1
                    if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                        if category_name in self.viewable_suites_by_category:
                            self.viewable_suites_by_category[category_name] += 1
                        else:
                            self.viewable_suites_by_category[category_name] = 1
                        if is_valid:
                            if category_name in self.viewable_valid_suites_by_category:
                                self.viewable_valid_suites_by_category[category_name] += 1
                            else:
                                self.viewable_valid_suites_by_category[category_name] = 1
                    if is_level_one_certified:
                        if category_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
                            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] += 1
                        else:
                            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 1
                        if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                            if category_name in self.certified_level_one_viewable_suites_by_category:
                                self.certified_level_one_viewable_suites_by_category[category_name] += 1
                            else:
                                self.certified_level_one_viewable_suites_by_category[category_name] = 1
                self.load_repository_and_suite_tuple(repository)
                if is_level_one_certified:
                    self.load_certified_level_one_repository_and_suite_tuple(repository)
        except Exception:
            # The viewable repository numbers and the categorized (filtered) lists of repository tuples
            # may be slightly skewed, but that is no reason to result in a potential server error.  All
            # will be corrected at next server start.
            log.exception("Handled error adding entry to repository registry")

    def edit_category_entry(self, old_name, new_name):
        if old_name in self.viewable_repositories_and_suites_by_category:
            val = self.viewable_repositories_and_suites_by_category[old_name]
            del self.viewable_repositories_and_suites_by_category[old_name]
            self.viewable_repositories_and_suites_by_category[new_name] = val
        else:
            self.viewable_repositories_and_suites_by_category[new_name] = 0
        if old_name in self.viewable_valid_repositories_and_suites_by_category:
            val = self.viewable_valid_repositories_and_suites_by_category[old_name]
            del self.viewable_valid_repositories_and_suites_by_category[old_name]
            self.viewable_valid_repositories_and_suites_by_category[new_name] = val
        else:
            self.viewable_valid_repositories_and_suites_by_category[new_name] = 0
        if old_name in self.viewable_suites_by_category:
            val = self.viewable_suites_by_category[old_name]
            del self.viewable_suites_by_category[old_name]
            self.viewable_suites_by_category[new_name] = val
        else:
            self.viewable_suites_by_category[new_name] = 0
        if old_name in self.viewable_valid_suites_by_category:
            val = self.viewable_valid_suites_by_category[old_name]
            del self.viewable_valid_suites_by_category[old_name]
            self.viewable_valid_suites_by_category[new_name] = val
        else:
            self.viewable_valid_suites_by_category[new_name] = 0
        if old_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
            val = self.certified_level_one_viewable_repositories_and_suites_by_category[old_name]
            del self.certified_level_one_viewable_repositories_and_suites_by_category[old_name]
            self.certified_level_one_viewable_repositories_and_suites_by_category[new_name] = val
        else:
            self.certified_level_one_viewable_repositories_and_suites_by_category[new_name] = 0
        if old_name in self.certified_level_one_viewable_suites_by_category:
            val = self.certified_level_one_viewable_suites_by_category[old_name]
            del self.certified_level_one_viewable_suites_by_category[old_name]
            self.certified_level_one_viewable_suites_by_category[new_name] = val
        else:
            self.certified_level_one_viewable_suites_by_category[new_name] = 0

    def get_certified_level_one_clause_list(self):
        certified_level_one_tuples = []
        clause_list = []
        for repository in self.sa_session.query(model.Repository).filter(
            and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false())
        ):
            certified_level_one_tuple = self.get_certified_level_one_tuple(repository)
            latest_installable_changeset_revision, is_level_one_certified = certified_level_one_tuple
            if is_level_one_certified:
                certified_level_one_tuples.append(certified_level_one_tuple)
                clause_list.append(
                    and_(
                        model.RepositoryMetadata.table.c.repository_id == repository.id,
                        model.RepositoryMetadata.table.c.changeset_revision == latest_installable_changeset_revision,
                    )
                )
        return clause_list

    def get_certified_level_one_tuple(self, repository):
        """
        Return True if the latest installable changeset_revision of the received repository is level one certified.
        """
        if repository is None:
            return (None, False)
        if repository.deleted or repository.deprecated:
            return (None, False)
        # Get the latest installable changeset revision since that is all that is currently configured for testing.
        latest_installable_changeset_revision = metadata_util.get_latest_downloadable_changeset_revision(
            self.app, repository
        )
        if latest_installable_changeset_revision not in [None, hg_util.INITIAL_CHANGELOG_HASH]:
            encoded_repository_id = self.app.security.encode_id(repository.id)
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                self.app, encoded_repository_id, latest_installable_changeset_revision
            )
            if repository_metadata:
                # No repository_metadata.
                return (latest_installable_changeset_revision, True)
        else:
            # No installable changeset_revision.
            return (None, False)

    def is_level_one_certified(self, repository_metadata):
        if repository_metadata:
            repository = repository_metadata.repository
            if repository:
                if repository.deprecated or repository.deleted:
                    return False
                tuple = (
                    str(repository.name),
                    str(repository.user.username),
                    str(repository_metadata.changeset_revision),
                )
                if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION]:
                    return tuple in self.certified_level_one_suite_tuples
                else:
                    return tuple in self.certified_level_one_repository_and_suite_tuples
        return False

    def is_valid(self, repository):
        if repository and not repository.deleted and not repository.deprecated and repository.downloadable_revisions:
            return True
        return False

    def load_certified_level_one_repository_and_suite_tuple(self, repository):
        # The received repository has been determined to be level one certified.
        name = str(repository.name)
        owner = str(repository.user.username)
        tip_changeset_hash = repository.tip()
        if tip_changeset_hash != hg_util.INITIAL_CHANGELOG_HASH:
            certified_level_one_tuple = (name, owner, tip_changeset_hash)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if certified_level_one_tuple not in self.certified_level_one_suite_tuples:
                    self.certified_level_one_suite_tuples.append(certified_level_one_tuple)
            else:
                if certified_level_one_tuple not in self.certified_level_one_repository_and_suite_tuples:
                    self.certified_level_one_repository_and_suite_tuples.append(certified_level_one_tuple)

    def load_repository_and_suite_tuple(self, repository):
        name = str(repository.name)
        owner = str(repository.user.username)
        for repository_metadata in repository.metadata_revisions:
            changeset_revision = str(repository_metadata.changeset_revision)
            tuple = (name, owner, changeset_revision)
            if tuple not in self.repository_and_suite_tuples:
                self.repository_and_suite_tuples.append(tuple)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if tuple not in self.suite_tuples:
                    self.suite_tuples.append(tuple)

    def load_repository_and_suite_tuples(self):
        # Load self.certified_level_one_repository_and_suite_tuples and self.certified_level_one_suite_tuples.
        for repository in (
            self.sa_session.query(model.Repository)
            .join(model.RepositoryMetadata.table)
            .filter(or_(*self.certified_level_one_clause_list))
            .join(model.User.table)
        ):
            self.load_certified_level_one_repository_and_suite_tuple(repository)
        # Load self.repository_and_suite_tuples and self.suite_tuples
        for repository in (
            self.sa_session.query(model.Repository)
            .filter(and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false()))
            .join(model.User.table)
        ):
            self.load_repository_and_suite_tuple(repository)

    def load_viewable_repositories_and_suites_by_category(self):
        # Clear all dictionaries just in case they were previously loaded.
        self.certified_level_one_viewable_repositories_and_suites_by_category = {}
        self.certified_level_one_viewable_suites_by_category = {}
        self.certified_level_two_viewable_repositories_and_suites_by_category = {}
        self.certified_level_two_viewable_suites_by_category = {}
        self.viewable_repositories_and_suites_by_category = {}
        self.viewable_suites_by_category = {}
        self.viewable_valid_repositories_and_suites_by_category = {}
        self.viewable_valid_suites_by_category = {}
        for category in self.sa_session.query(model.Category):
            category_name = str(category.name)
            if category not in self.certified_level_one_viewable_repositories_and_suites_by_category:
                self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 0
            if category not in self.certified_level_one_viewable_suites_by_category:
                self.certified_level_one_viewable_suites_by_category[category_name] = 0
            if category not in self.viewable_repositories_and_suites_by_category:
                self.viewable_repositories_and_suites_by_category[category_name] = 0
            if category not in self.viewable_suites_by_category:
                self.viewable_suites_by_category[category_name] = 0
            if category not in self.viewable_valid_repositories_and_suites_by_category:
                self.viewable_valid_repositories_and_suites_by_category[category_name] = 0
            if category not in self.viewable_valid_suites_by_category:
                self.viewable_valid_suites_by_category[category_name] = 0
            for rca in category.repositories:
                repository = rca.repository
                if not repository.deleted and not repository.deprecated:
                    is_valid = self.is_valid(repository)
                    encoded_repository_id = self.app.security.encode_id(repository.id)
                    tip_changeset_hash = repository.tip()
                    repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                        self.app, encoded_repository_id, tip_changeset_hash
                    )
                    self.viewable_repositories_and_suites_by_category[category_name] += 1
                    if is_valid:
                        self.viewable_valid_repositories_and_suites_by_category[category_name] += 1
                    if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION]:
                        self.viewable_suites_by_category[category_name] += 1
                        if is_valid:
                            self.viewable_valid_suites_by_category[category_name] += 1
                    if self.is_level_one_certified(repository_metadata):
                        self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] += 1
                        if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION]:
                            self.certified_level_one_viewable_suites_by_category[category_name] += 1

    def remove_category_entry(self, category):
        catgeory_name = str(category.name)
        if catgeory_name in self.viewable_repositories_and_suites_by_category:
            del self.viewable_repositories_and_suites_by_category[catgeory_name]
        if catgeory_name in self.viewable_valid_repositories_and_suites_by_category:
            del self.viewable_valid_repositories_and_suites_by_category[catgeory_name]
        if catgeory_name in self.viewable_suites_by_category:
            del self.viewable_suites_by_category[catgeory_name]
        if catgeory_name in self.viewable_valid_suites_by_category:
            del self.viewable_valid_suites_by_category[catgeory_name]
        if catgeory_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
            del self.certified_level_one_viewable_repositories_and_suites_by_category[catgeory_name]
        if catgeory_name in self.certified_level_one_viewable_suites_by_category:
            del self.certified_level_one_viewable_suites_by_category[catgeory_name]

    def remove_entry(self, repository):
        try:
            if repository:
                is_valid = self.is_valid(repository)
                certified_level_one_tuple = self.get_certified_level_one_tuple(repository)
                latest_installable_changeset_revision, is_level_one_certified = certified_level_one_tuple
                for rca in repository.categories:
                    category = rca.category
                    category_name = str(category.name)
                    if category_name in self.viewable_repositories_and_suites_by_category:
                        if self.viewable_repositories_and_suites_by_category[category_name] > 0:
                            self.viewable_repositories_and_suites_by_category[category_name] -= 1
                    else:
                        self.viewable_repositories_and_suites_by_category[category_name] = 0
                    if is_valid:
                        if category_name in self.viewable_valid_repositories_and_suites_by_category:
                            if self.viewable_valid_repositories_and_suites_by_category[category_name] > 0:
                                self.viewable_valid_repositories_and_suites_by_category[category_name] -= 1
                        else:
                            self.viewable_valid_repositories_and_suites_by_category[category_name] = 0
                    if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                        if category_name in self.viewable_suites_by_category:
                            if self.viewable_suites_by_category[category_name] > 0:
                                self.viewable_suites_by_category[category_name] -= 1
                        else:
                            self.viewable_suites_by_category[category_name] = 0
                        if is_valid:
                            if category_name in self.viewable_valid_suites_by_category:
                                if self.viewable_valid_suites_by_category[category_name] > 0:
                                    self.viewable_valid_suites_by_category[category_name] -= 1
                            else:
                                self.viewable_valid_suites_by_category[category_name] = 0
                    if is_level_one_certified:
                        if category_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
                            if self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] > 0:
                                self.certified_level_one_viewable_repositories_and_suites_by_category[
                                    category_name
                                ] -= 1
                        else:
                            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 0
                        if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                            if category_name in self.certified_level_one_viewable_suites_by_category:
                                if self.certified_level_one_viewable_suites_by_category[category_name] > 0:
                                    self.certified_level_one_viewable_suites_by_category[category_name] -= 1
                            else:
                                self.certified_level_one_viewable_suites_by_category[category_name] = 0
                self.unload_repository_and_suite_tuple(repository)
                if is_level_one_certified:
                    self.unload_certified_level_one_repository_and_suite_tuple(repository)
        except Exception:
            # The viewable repository numbers and the categorized (filtered) lists of repository tuples
            # may be slightly skewed, but that is no reason to result in a potential server error.  All
            # will be corrected at next server start.
            log.exception("Handled error removing entry from repository registry")

    @property
    def sa_session(self):
        return self.app.model.session

    def unload_certified_level_one_repository_and_suite_tuple(self, repository):
        # The received repository has been determined to be level one certified.
        name = str(repository.name)
        owner = str(repository.user.username)
        tip_changeset_hash = repository.tip()
        if tip_changeset_hash != hg_util.INITIAL_CHANGELOG_HASH:
            certified_level_one_tuple = (name, owner, tip_changeset_hash)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if certified_level_one_tuple in self.certified_level_one_suite_tuples:
                    self.certified_level_one_suite_tuples.remove(certified_level_one_tuple)
            else:
                if certified_level_one_tuple in self.certified_level_one_repository_and_suite_tuples:
                    self.certified_level_one_repository_and_suite_tuples.remove(certified_level_one_tuple)

    def unload_repository_and_suite_tuple(self, repository):
        name = str(repository.name)
        owner = str(repository.user.username)
        for repository_metadata in repository.metadata_revisions:
            changeset_revision = str(repository_metadata.changeset_revision)
            tuple = (name, owner, changeset_revision)
            if tuple in self.repository_and_suite_tuples:
                self.repository_and_suite_tuples.remove(tuple)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if tuple in self.suite_tuples:
                    self.suite_tuples.remove(tuple)



import logging
from typing import Optional

from sqlalchemy import (
    false,
    true,
)

from galaxy import (
    model,
    util,
)
from galaxy.managers.context import ProvidesUserContext
from galaxy.managers.quotas import QuotaManager
from galaxy.quota._schema import (
    CreateQuotaParams,
    CreateQuotaResult,
    DefaultQuotaValues,
    DeleteQuotaPayload,
    QuotaDetails,
    QuotaSummaryList,
    UpdateQuotaParams,
)
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.web import url_for
from galaxy.webapps.galaxy.services.base import ServiceBase

log = logging.getLogger(__name__)


class QuotasService(ServiceBase):
    """Interface/service object shared by controllers for interacting with quotas."""

    def __init__(self, security: IdEncodingHelper, quota_manager: QuotaManager):
        super().__init__(security)
        self.quota_manager = quota_manager

    def index(self, trans: ProvidesUserContext, deleted: bool = False) -> QuotaSummaryList:
        """Displays a list of quotas."""
        rval = []
        query = trans.sa_session.query(model.Quota)
        if deleted:
            route = "deleted_quota"
            query = query.filter(model.Quota.deleted == true())
        else:
            route = "quota"
            query = query.filter(model.Quota.deleted == false())
        for quota in query:
            item = quota.to_dict(value_mapper={"id": DecodedDatabaseIdField.encode})
            encoded_id = DecodedDatabaseIdField.encode(quota.id)
            item["url"] = url_for(route, id=encoded_id)
            rval.append(item)
        return QuotaSummaryList.construct(__root__=rval)

    def show(self, trans: ProvidesUserContext, id: DecodedDatabaseIdField, deleted: bool = False) -> QuotaDetails:
        """Displays information about a quota."""
        quota = self.quota_manager.get_quota(trans, id, deleted=deleted)
        rval = quota.to_dict(
            view="element", value_mapper={"id": DecodedDatabaseIdField.encode, "total_disk_usage": float}
        )
        return QuotaDetails.construct(**rval)

    def create(self, trans: ProvidesUserContext, params: CreateQuotaParams) -> CreateQuotaResult:
        """Creates a new quota."""
        payload = params.dict()
        self.validate_in_users_and_groups(trans, payload)
        quota, message = self.quota_manager.create_quota(payload)
        item = quota.to_dict(value_mapper={"id": DecodedDatabaseIdField.encode})
        item["url"] = url_for("quota", id=DecodedDatabaseIdField.encode(quota.id))
        item["message"] = message
        return CreateQuotaResult.construct(**item)

    def update(self, trans: ProvidesUserContext, id: DecodedDatabaseIdField, params: UpdateQuotaParams) -> str:
        """Modifies a quota."""
        payload = params.dict()
        self.validate_in_users_and_groups(trans, payload)
        quota = self.quota_manager.get_quota(trans, id, deleted=False)

        params = UpdateQuotaParams(**payload)
        # FIXME: Doing it this way makes the update non-atomic if a method fails after an earlier one has succeeded.
        methods = []
        if params.name or params.description:
            methods.append(self.quota_manager.rename_quota)
        if params.amount:
            methods.append(self.quota_manager.edit_quota)
        if params.default == DefaultQuotaValues.NO:
            methods.append(self.quota_manager.unset_quota_default)
        elif params.default:
            methods.append(self.quota_manager.set_quota_default)
        if params.in_users or params.in_groups:
            methods.append(self.quota_manager.manage_users_and_groups_for_quota)

        messages = []
        for method in methods:
            message = method(quota, params)
            messages.append(message)
        return "; ".join(messages)

    def delete(
        self, trans: ProvidesUserContext, id: DecodedDatabaseIdField, payload: Optional[DeleteQuotaPayload] = None
    ) -> str:
        """Marks a quota as deleted."""
        quota = self.quota_manager.get_quota(
            trans, id, deleted=False
        )  # deleted quotas are not technically members of this collection
        message = self.quota_manager.delete_quota(quota)
        if payload and payload.purge:
            message += self.quota_manager.purge_quota(quota)
        return message

    def undelete(self, trans: ProvidesUserContext, id: DecodedDatabaseIdField) -> str:
        """Restores a previously deleted quota."""
        quota = self.quota_manager.get_quota(trans, id, deleted=True)
        return self.quota_manager.undelete_quota(quota)

    def validate_in_users_and_groups(self, trans, payload):
        """
        For convenience, in_users and in_groups can be encoded IDs or emails/group names in the API.
        """

        def get_id(item, model_class, column):
            try:
                return trans.security.decode_id(item)
            except Exception:
                pass  # maybe an email/group name
            # this will raise if the item is invalid
            return trans.sa_session.query(model_class).filter(column == item).first().id

        new_in_users = []
        new_in_groups = []
        invalid = []
        for item in util.listify(payload.get("in_users", [])):
            try:
                new_in_users.append(get_id(item, model.User, model.User.email))
            except Exception:
                invalid.append(item)
        for item in util.listify(payload.get("in_groups", [])):
            try:
                new_in_groups.append(get_id(item, model.Group, model.Group.name))
            except Exception:
                invalid.append(item)
        if invalid:
            msg = (
                f"The following value(s) for associated users and/or groups could not be parsed: {', '.join(invalid)}."
            )
            msg += "  Valid values are email addresses of users, names of groups, or IDs of both."
            raise Exception(msg)
        payload["in_users"] = list(map(str, new_in_users))
        payload["in_groups"] = list(map(str, new_in_groups))



import logging

from galaxy.managers.notification import NotificationManager
from galaxy.managers.visualizations import (
    VisualizationManager,
    VisualizationSerializer,
)
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.webapps.galaxy.services.base import ServiceBase
from galaxy.webapps.galaxy.services.sharable import ShareableService

log = logging.getLogger(__name__)


class VisualizationsService(ServiceBase):
    """Common interface/service logic for interactions with visualizations in the context of the API.

    Provides the logic of the actions invoked by API controllers and uses type definitions
    and pydantic models to declare its parameters and return types.
    """

    def __init__(
        self,
        security: IdEncodingHelper,
        manager: VisualizationManager,
        serializer: VisualizationSerializer,
        notification_manager: NotificationManager,
    ):
        super().__init__(security)
        self.manager = manager
        self.serializer = serializer
        self.shareable_service = ShareableService(self.manager, self.serializer, notification_manager)

    # TODO: add the rest of the API actions here and call them directly from the API controller



"""
API operations on FormDefinition objects.
"""
import logging

from galaxy import web
from galaxy.forms.forms import form_factory
from galaxy.model.base import transaction
from galaxy.util import XML
from galaxy.webapps.base.controller import url_for
from . import BaseGalaxyAPIController

log = logging.getLogger(__name__)


class FormDefinitionAPIController(BaseGalaxyAPIController):
    @web.legacy_expose_api
    def index(self, trans, **kwd):
        """
        GET /api/forms
        Displays a collection (list) of forms.
        """
        if not trans.user_is_admin:
            trans.response.status = 403
            return "You are not authorized to view the list of forms."
        query = trans.sa_session.query(trans.app.model.FormDefinition)
        rval = []
        for form_definition in query:
            item = form_definition.to_dict(
                value_mapper={"id": trans.security.encode_id, "form_definition_current_id": trans.security.encode_id}
            )
            item["url"] = url_for("form", id=trans.security.encode_id(form_definition.id))
            rval.append(item)
        return rval

    @web.legacy_expose_api
    def show(self, trans, id, **kwd):
        """
        GET /api/forms/{encoded_form_id}
        Displays information about a form.
        """
        form_definition_id = id
        try:
            decoded_form_definition_id = trans.security.decode_id(form_definition_id)
        except TypeError:
            trans.response.status = 400
            return f"Malformed form definition id ( {str(form_definition_id)} ) specified, unable to decode."
        try:
            form_definition = trans.sa_session.query(trans.app.model.FormDefinition).get(decoded_form_definition_id)
        except Exception:
            form_definition = None
        if not form_definition or not trans.user_is_admin:
            trans.response.status = 400
            return f"Invalid form definition id ( {str(form_definition_id)} ) specified."
        item = form_definition.to_dict(
            view="element",
            value_mapper={"id": trans.security.encode_id, "form_definition_current_id": trans.security.encode_id},
        )
        item["url"] = url_for("form", id=form_definition_id)
        return item

    @web.legacy_expose_api
    def create(self, trans, payload, **kwd):
        """
        POST /api/forms
        Creates a new form.
        """
        if not trans.user_is_admin:
            trans.response.status = 403
            return "You are not authorized to create a new form."
        xml_text = payload.get("xml_text", None)
        if xml_text is None:
            trans.response.status = 400
            return "Missing required parameter 'xml_text'."
            # enhance to allow creating from more than just xml
        form_definition = form_factory.from_elem(XML(xml_text))
        trans.sa_session.add(form_definition)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        encoded_id = trans.security.encode_id(form_definition.id)
        item = form_definition.to_dict(
            view="element",
            value_mapper={"id": trans.security.encode_id, "form_definition_current_id": trans.security.encode_id},
        )
        item["url"] = url_for("form", id=encoded_id)
        return [item]



import logging

from galaxy import (
    util,
    web,
)
from galaxy.exceptions import ObjectNotFound
from galaxy.web import (
    expose_api,
    expose_api_anonymous_and_sessionless,
)
from galaxy.webapps.base.controller import BaseAPIController

log = logging.getLogger(__name__)


class DynamicToolsController(BaseAPIController):
    """
    RESTful controller for interactions with dynamic tools.

    Dynamic tools are tools defined in the database. Use the tools controller
    to run these tools and view functional information.
    """

    @expose_api_anonymous_and_sessionless
    def index(self, trans, **kwds):
        """
        GET /api/dynamic_tools

        This returns meta-information about the dynamic tool, such as
        tool_uuid. To use the tool or view funtional information such as
        inputs and outputs, use the standard tools API indexed by the
        ID (and optionally version) returned from this endpoint.
        """
        manager = self.app.dynamic_tools_manager
        return list(map(lambda t: t.to_dict(), manager.list_tools()))

    @expose_api_anonymous_and_sessionless
    def show(self, trans, id, **kwd):
        """
        GET /api/dynamic_tools/{encoded_dynamic_tool_id|tool_uuid}
        """
        self._get_dynamic_tool(trans, id).to_dict()

    @web.require_admin
    @expose_api
    def create(self, trans, payload, **kwd):
        """
        POST /api/dynamic_tools

        The payload is expected to be a tool definition to dynamically load
        into Galaxy's toolbox.

        :type representation: dict
        :param representation: a JSON-ified tool description to load
        :type uuid: str
        :param uuid: the uuid to associate with the tool being created
        """
        dynamic_tool = self.app.dynamic_tools_manager.create_tool(
            trans, payload, allow_load=util.asbool(kwd.get("allow_load", True))
        )
        return dynamic_tool.to_dict()

    @web.require_admin
    @expose_api
    def delete(self, trans, id, **kwd):
        """
        DELETE /api/dynamic_tools/{encoded_dynamic_tool_id|tool_uuid}

        Deactivate the specified dynamic tool. Deactivated tools will not
        be loaded into the toolbox.
        """
        manager = self.app.dynamic_tools_manager
        dynamic_tool = self._get_dynamic_tool(trans, id)
        updated_dynamic_tool = manager.deactivate(dynamic_tool)
        return updated_dynamic_tool.to_dict()

    def _get_dynamic_tool(self, trans, request_id):
        manager = self.app.dynamic_tools_manager
        if util.is_uuid(request_id):
            dynamic_tool = manager.get_tool_by_uuid(request_id)
        else:
            dynamic_tool = manager.get_tool_by_id(trans.security.decode_id(request_id))
        if dynamic_tool is None:
            raise ObjectNotFound()
        return dynamic_tool



"""
API operations related to tagging items.
"""
import logging

from galaxy import exceptions
from galaxy.web import expose_api
from galaxy.webapps.base.controller import UsesTagsMixin
from . import BaseGalaxyAPIController

log = logging.getLogger(__name__)


class BaseItemTagsController(BaseGalaxyAPIController, UsesTagsMixin):
    """ """

    @expose_api
    def index(self, trans, **kwd):
        """ """
        tags = self._get_user_tags(trans, self.tagged_item_class, kwd[self.tagged_item_id])
        return [self._api_value(tag, trans, view="collection") for tag in tags]

    @expose_api
    def show(self, trans, tag_name, **kwd):
        """ """
        tag = self._get_item_tag_assoc(trans, self.tagged_item_class, kwd[self.tagged_item_id], tag_name)
        if not tag:
            raise exceptions.ObjectNotFound("Failed to retrieve specified tag.")
        return self._api_value(tag, trans)

    @expose_api
    def create(self, trans, tag_name, payload=None, **kwd):
        """ """
        payload = payload or {}
        value = payload.get("value", None)
        tag = self._apply_item_tag(trans, self.tagged_item_class, kwd[self.tagged_item_id], tag_name, value)
        return self._api_value(tag, trans)

    # Not handling these differently at this time
    update = create

    @expose_api
    def delete(self, trans, tag_name, **kwd):
        """ """
        deleted = self._remove_items_tag(trans, self.tagged_item_class, kwd[self.tagged_item_id], tag_name)
        if not deleted:
            raise exceptions.RequestParameterInvalidException("Failed to delete specified tag.")
        # TODO: ugh - 204 would be better
        return "OK"

    def _api_value(self, tag, trans, view="element"):
        return tag.to_dict(view=view, value_mapper={"id": trans.security.encode_id})


class HistoryContentTagsController(BaseItemTagsController):
    controller_name = "history_content_tags"
    tagged_item_class = "HistoryDatasetAssociation"
    tagged_item_id = "history_content_id"


class HistoryTagsController(BaseItemTagsController):
    controller_name = "history_tags"
    tagged_item_class = "History"
    tagged_item_id = "history_id"


class WorkflowTagsController(BaseItemTagsController):
    controller_name = "workflow_tags"
    tagged_item_class = "StoredWorkflow"
    tagged_item_id = "workflow_id"


# TODO: Visualization and Pages once APIs for those are available



"""
API operations on defining cloud authorizations.

Through means of cloud authorization a user is able to grant a Galaxy server a secure access to his/her
cloud-based resources without sharing his/her long-lasting credentials.

User provides a provider-specific configuration, which Galaxy users to request temporary credentials
from the provider to access the user's resources.
"""

import logging

from galaxy.exceptions import (
    ActionInputError,
    InternalServerError,
    MalformedId,
    RequestParameterInvalidException,
    RequestParameterMissingException,
)
from galaxy.managers import cloudauthzs
from galaxy.model.base import transaction
from galaxy.structured_app import StructuredApp
from galaxy.util import unicodify
from galaxy.web import expose_api
from . import BaseGalaxyAPIController

log = logging.getLogger(__name__)


class CloudAuthzController(BaseGalaxyAPIController):
    """
    RESTfull controller for defining cloud authorizations.
    """

    def __init__(self, app: StructuredApp):
        super().__init__(app)
        self.cloudauthz_manager = cloudauthzs.CloudAuthzManager(app)
        self.cloudauthz_serializer = cloudauthzs.CloudAuthzsSerializer(app)
        self.cloudauthz_deserializer = cloudauthzs.CloudAuthzsDeserializer(app)

    @expose_api
    def index(self, trans, **kwargs):
        """
        GET /api/cloud/authz

        Lists all the cloud authorizations user has defined.

        :type  trans: galaxy.webapps.base.webapp.GalaxyWebTransaction
        :param trans: Galaxy web transaction

        :param kwargs: empty dict

        :rtype: list of dict
        :return: a list of cloud authorizations (each represented in key-value pair format) defined for the user.
        """
        rtv = []
        for cloudauthz in trans.user.cloudauthz:
            rtv.append(
                self.cloudauthz_serializer.serialize_to_view(
                    cloudauthz, user=trans.user, trans=trans, **self._parse_serialization_params(kwargs, "summary")
                )
            )
        return rtv

    @expose_api
    def create(self, trans, payload, **kwargs):
        """
        * POST /api/cloud/authz
            Request to store the payload as a cloudauthz (cloud authorization) configuration for a user.

        :type  trans: galaxy.webapps.base.webapp.GalaxyWebTransaction
        :param trans: Galaxy web transaction

        :type payload: dict
        :param payload: A dictionary structure containing the following keys:
            *   provider:       the cloud-based resource provider to which this configuration belongs to.

            *   config:         a dictionary containing all the configuration required to request temporary credentials
                                from the provider. See the following page for details:
                                https://galaxyproject.org/authnz/

            *   authn_id:       the (encoded) ID of a third-party authentication of a user. To have this ID, user must
                                have logged-in to this Galaxy server using third-party identity (e.g., Google), or has
                                associated his/her Galaxy account with a third-party OIDC-based identity. See this page:
                                https://galaxyproject.org/authnz/config/

            *   description:    [Optional] a brief description for this configuration.

        :param kwargs: empty dict

        :rtype: dict
        :return: a dictionary with the following kvp:
            *   status:     HTTP response code
            *   message:    A message complementary to the response code.
        """
        msg_template = f"Rejected user `{str(trans.user.id)}`'s request to create cloudauthz config because of {{}}."
        if not isinstance(payload, dict):
            raise ActionInputError(
                "Invalid payload data type. The payload is expected to be a dictionary, but "
                "received data of type `{}`.".format(str(type(payload)))
            )

        missing_arguments = []
        provider = payload.get("provider", None)
        if provider is None:
            missing_arguments.append("provider")

        config = payload.get("config", None)
        if config is None:
            missing_arguments.append("config")

        authn_id = payload.get("authn_id", None)
        if authn_id is None and provider.lower() not in ["azure", "gcp"]:
            missing_arguments.append("authn_id")

        if len(missing_arguments) > 0:
            log.debug(msg_template.format(f"missing required config {missing_arguments}"))
            raise RequestParameterMissingException(
                "The following required arguments are missing in the payload: " "{}".format(missing_arguments)
            )

        description = payload.get("description", "")

        if not isinstance(config, dict):
            log.debug(msg_template.format(f"invalid config type `{type(config)}`, expect `dict`"))
            raise RequestParameterInvalidException(
                "Invalid type for the required `config` variable; expect `dict` "
                "but received `{}`.".format(type(config))
            )
        if authn_id:
            try:
                decoded_authn_id = self.decode_id(authn_id)
            except MalformedId as e:
                log.debug(msg_template.format(f"cannot decode authz_id `{authn_id}`"))
                raise e

            try:
                trans.app.authnz_manager.can_user_assume_authn(trans, decoded_authn_id)
            except Exception as e:
                raise e

        # No two authorization configuration with
        # exact same key/value should exist.
        for ca in trans.user.cloudauthz:
            if ca.equals(trans.user.id, provider, authn_id, config):
                log.debug(
                    "Rejected user `{}`'s request to create cloud authorization because a similar config "
                    "already exists.".format(trans.user.id)
                )
                raise ActionInputError("A similar cloud authorization configuration is already defined.")

        try:
            new_cloudauthz = self.cloudauthz_manager.create(
                user_id=trans.user.id, provider=provider, config=config, authn_id=authn_id, description=description
            )
            view = self.cloudauthz_serializer.serialize_to_view(
                new_cloudauthz, trans=trans, **self._parse_serialization_params(kwargs, "summary")
            )
            log.debug(f"Created a new cloudauthz record for the user id `{str(trans.user.id)}` ")
            return view
        except Exception as e:
            log.exception(msg_template.format("exception while creating the new cloudauthz record"))
            raise InternalServerError(
                "An unexpected error has occurred while responding to the create request of the "
                "cloudauthz API." + unicodify(e)
            )

    @expose_api
    def delete(self, trans, encoded_authz_id, **kwargs):
        """
        * DELETE /api/cloud/authz/{encoded_authz_id}
            Deletes the CloudAuthz record with the given ``encoded_authz_id`` from database.

        :type  trans: galaxy.webapps.base.webapp.GalaxyWebTransaction
        :param trans: Galaxy web transaction

        :type  encoded_authz_id:    string
        :param encoded_authz_id:    The encoded ID of the CloudAuthz record to be marked deleted.

        :rtype  JSON
        :return The cloudauthz record marked as deleted, serialized as a JSON object.
        """

        msg_template = f"Rejected user `{str(trans.user.id)}`'s request to delete cloudauthz config because of {{}}."
        try:
            authz_id = self.decode_id(encoded_authz_id)
        except MalformedId as e:
            log.debug(msg_template.format(f"cannot decode authz_id `{encoded_authz_id}`"))
            raise e

        try:
            cloudauthz = trans.app.authnz_manager.try_get_authz_config(trans.sa_session, trans.user.id, authz_id)
            trans.sa_session.delete(cloudauthz)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            log.debug(f"Deleted a cloudauthz record with id `{authz_id}` for the user id `{str(trans.user.id)}` ")
            view = self.cloudauthz_serializer.serialize_to_view(
                cloudauthz, trans=trans, **self._parse_serialization_params(kwargs, "summary")
            )
            trans.response.status = "200"
            return view
        except Exception as e:
            log.exception(
                msg_template.format(
                    "exception while deleting the cloudauthz record with " "ID: `{}`.".format(encoded_authz_id)
                )
            )
            raise InternalServerError(
                "An unexpected error has occurred while responding to the DELETE request of the "
                "cloudauthz API." + unicodify(e)
            )

    @expose_api
    def update(self, trans, encoded_authz_id, payload, **kwargs):
        """
        PUT /api/cloud/authz/{encoded_authz_id}

        Updates the values for the cloudauthz configuration with the given ``encoded_authz_id``.

        With this API only the following attributes of a cloudauthz configuration
        can be updated: `authn_id`, `provider`, `config`, `deleted`.

        :type  trans:               galaxy.webapps.base.webapp.GalaxyWebTransaction
        :param trans:               Galaxy web transaction

        :type  encoded_authz_id:    string
        :param encoded_authz_id:    The encoded ID of the CloudAuthz record to be updated.

        :type payload:              dict
        :param payload:             A dictionary structure containing the attributes to modified with their new values.
                                    It can contain any number of the following attributes:

                                        *   provider:   the cloud-based resource provider
                                                        to which this configuration belongs to.

                                        *   authn_id:   the (encoded) ID of a third-party authentication of a user.
                                                        To have this ID, user must have logged-in to this Galaxy server
                                                        using third-party identity (e.g., Google), or has associated
                                                        their Galaxy account with a third-party OIDC-based identity.
                                                        See this page: https://galaxyproject.org/authnz/config/

                                                        Note: A user can associate a cloudauthz record with their own
                                                        authentications only. If the given authentication with authn_id
                                                        belongs to a different user, Galaxy will throw the
                                                        ItemAccessibilityException exception.

                                        *   config:     a dictionary containing all the configuration required to
                                                        request temporary credentials from the provider.
                                                        See the following page for details:
                                                        https://galaxyproject.org/authnz/

                                        *   deleted:    a boolean type marking the specified cloudauthz as (un)deleted.

        """

        msg_template = f"Rejected user `{str(trans.user.id)}`'s request to delete cloudauthz config because of {{}}."
        try:
            authz_id = self.decode_id(encoded_authz_id)
        except MalformedId as e:
            log.debug(msg_template.format(f"cannot decode authz_id `{encoded_authz_id}`"))
            raise e

        try:
            cloudauthz_to_update = trans.app.authnz_manager.try_get_authz_config(
                trans.sa_session, trans.user.id, authz_id
            )
            self.cloudauthz_deserializer.deserialize(cloudauthz_to_update, payload, trans=trans)
            self.cloudauthz_serializer.serialize_to_view(cloudauthz_to_update, view="summary")
            return self.cloudauthz_serializer.serialize_to_view(cloudauthz_to_update, view="summary")
        except MalformedId as e:
            raise e
        except Exception as e:
            log.exception(
                msg_template.format(
                    "exception while updating the cloudauthz record with " "ID: `{}`.".format(encoded_authz_id)
                )
            )
            raise InternalServerError(
                "An unexpected error has occurred while responding to the PUT request of the "
                "cloudauthz API." + unicodify(e)
            )



from markupsafe import escape
from sqlalchemy import (
    desc,
    false,
    true,
)
from sqlalchemy.orm import (
    joinedload,
    undefer,
)

from galaxy import (
    model,
    util,
    web,
)
from galaxy.managers.hdas import HDAManager
from galaxy.managers.histories import (
    HistoryManager,
    HistorySerializer,
)
from galaxy.managers.pages import PageManager
from galaxy.managers.sharable import SlugBuilder
from galaxy.managers.workflows import WorkflowsManager
from galaxy.model.base import transaction
from galaxy.model.item_attrs import UsesItemRatings
from galaxy.schema.schema import CreatePagePayload
from galaxy.structured_app import StructuredApp
from galaxy.util.sanitize_html import sanitize_html
from galaxy.web import (
    error,
    url_for,
)
from galaxy.web.framework.helpers import (
    grids,
    time_ago,
)
from galaxy.webapps.base.controller import (
    BaseUIController,
    SharableMixin,
    UsesStoredWorkflowMixin,
    UsesVisualizationMixin,
)
from galaxy.webapps.galaxy.api import depends


def format_bool(b):
    if b:
        return "yes"
    else:
        return ""


class PageListGrid(grids.Grid):
    # Custom column.
    class URLColumn(grids.PublicURLColumn):
        def get_value(self, trans, grid, item):
            return url_for(
                controller="page", action="display_by_username_and_slug", username=item.user.username, slug=item.slug
            )

    # Grid definition
    use_panels = True
    title = "Pages"
    model_class = model.Page
    default_filter = {"published": "All", "tags": "All", "title": "All", "sharing": "All"}
    default_sort_key = "-update_time"
    columns = [
        grids.TextColumn(
            "Title",
            key="title",
            attach_popup=True,
            filterable="advanced",
            link=(
                lambda item: dict(action="display_by_username_and_slug", username=item.user.username, slug=item.slug)
            ),
        ),
        URLColumn("Permalink"),
        grids.OwnerAnnotationColumn(
            "Annotation",
            key="annotation",
            model_annotation_association_class=model.PageAnnotationAssociation,
            filterable="advanced",
        ),
        grids.IndividualTagsColumn(
            "Tags",
            key="tags",
            model_tag_association_class=model.PageTagAssociation,
            filterable="advanced",
            grid_name="PageListGrid",
        ),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False),
        grids.GridColumn("Created", key="create_time", format=time_ago),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    global_actions = [grids.GridAction("Add new page", dict(controller="", action="pages/create"))]
    operations = [
        grids.DisplayByUsernameAndSlugGridOperation("View", allow_multiple=False),
        grids.GridOperation("Edit content", allow_multiple=False, url_args=dict(controller="", action="pages/editor")),
        grids.GridOperation("Edit attributes", allow_multiple=False, url_args=dict(controller="", action="pages/edit")),
        grids.GridOperation(
            "Share or Publish",
            allow_multiple=False,
            condition=(lambda item: not item.deleted),
            url_args=dict(controller="", action="pages/sharing"),
        ),
        grids.GridOperation("Delete", confirm="Are you sure you want to delete this page?"),
    ]

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter_by(user=trans.user, deleted=False)


class PageAllPublishedGrid(grids.Grid):
    # Grid definition
    use_panels = True
    title = "Published Pages"
    model_class = model.Page
    default_sort_key = "update_time"
    default_filter = dict(title="All", username="All")
    columns = [
        grids.PublicURLColumn("Title", key="title", filterable="advanced"),
        grids.OwnerAnnotationColumn(
            "Annotation",
            key="annotation",
            model_annotation_association_class=model.PageAnnotationAssociation,
            filterable="advanced",
        ),
        grids.OwnerColumn("Owner", key="username", model_class=model.User, filterable="advanced"),
        grids.CommunityRatingColumn("Community Rating", key="rating"),
        grids.CommunityTagsColumn(
            "Community Tags",
            key="tags",
            model_tag_association_class=model.PageTagAssociation,
            filterable="advanced",
            grid_name="PageAllPublishedGrid",
        ),
        grids.ReverseSortColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search title, annotation, owner, and tags",
            cols_to_filter=[columns[0], columns[1], columns[2], columns[4]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwargs):
        # See optimization description comments and TODO for tags in matching public histories query.
        return (
            trans.sa_session.query(self.model_class)
            .join("user")
            .filter(model.User.deleted == false())
            .options(
                joinedload(self.model_class.user).load_only("username"),
                joinedload(self.model_class.annotations),
                undefer("average_rating"),
            )
        )

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter(self.model_class.deleted == false()).filter(self.model_class.published == true())


class ItemSelectionGrid(grids.Grid):
    """Base class for pages' item selection grids."""

    # Custom columns.
    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, item):
            if hasattr(item, "get_display_name"):
                return escape(item.get_display_name())
            else:
                return escape(item.name)

    # Grid definition.
    show_item_checkboxes = True
    default_filter = {"deleted": "False", "sharing": "All"}
    default_sort_key = "-update_time"
    use_paging = True
    num_rows_per_page = 10

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter_by(user=trans.user)


class HistorySelectionGrid(ItemSelectionGrid):
    """Grid for selecting histories."""

    # Grid definition.
    title = "Saved Histories"
    model_class = model.History
    columns = [
        ItemSelectionGrid.NameColumn("Name", key="name", filterable="advanced"),
        grids.IndividualTagsColumn(
            "Tags", key="tags", model_tag_association_class=model.HistoryTagAssociation, filterable="advanced"
        ),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False, visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter_by(user=trans.user, purged=False)


class HistoryDatasetAssociationSelectionGrid(ItemSelectionGrid):
    """Grid for selecting HDAs."""

    # Grid definition.
    title = "Saved Datasets"
    model_class = model.HistoryDatasetAssociation
    columns = [
        ItemSelectionGrid.NameColumn("Name", key="name", filterable="advanced"),
        grids.IndividualTagsColumn(
            "Tags",
            key="tags",
            model_tag_association_class=model.HistoryDatasetAssociationTagAssociation,
            filterable="advanced",
        ),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False, visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def apply_query_filter(self, trans, query, **kwargs):
        # To filter HDAs by user, need to join HDA and History table and then filter histories by user. This is necessary because HDAs do not have
        # a user relation.
        return query.select_from(model.HistoryDatasetAssociation.table.join(model.History.table)).filter(
            model.History.user == trans.user
        )


class WorkflowSelectionGrid(ItemSelectionGrid):
    """Grid for selecting workflows."""

    # Grid definition.
    title = "Saved Workflows"
    model_class = model.StoredWorkflow
    columns = [
        ItemSelectionGrid.NameColumn("Name", key="name", filterable="advanced"),
        grids.IndividualTagsColumn(
            "Tags", key="tags", model_tag_association_class=model.StoredWorkflowTagAssociation, filterable="advanced"
        ),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False, visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )


class PageSelectionGrid(ItemSelectionGrid):
    """Grid for selecting pages."""

    # Grid definition.
    title = "Saved Pages"
    model_class = model.Page
    columns = [
        grids.TextColumn("Title", key="title", filterable="advanced"),
        grids.IndividualTagsColumn(
            "Tags", key="tags", model_tag_association_class=model.PageTagAssociation, filterable="advanced"
        ),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False, visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )


class VisualizationSelectionGrid(ItemSelectionGrid):
    """Grid for selecting visualizations."""

    # Grid definition.
    title = "Saved Visualizations"
    model_class = model.Visualization
    columns = [
        grids.TextColumn("Title", key="title", filterable="advanced"),
        grids.TextColumn("Type", key="type"),
        grids.IndividualTagsColumn(
            "Tags",
            key="tags",
            model_tag_association_class=model.VisualizationTagAssociation,
            filterable="advanced",
            grid_name="VisualizationListGrid",
        ),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )


# Adapted from the _BaseHTMLProcessor class of https://github.com/kurtmckee/feedparser
class PageController(BaseUIController, SharableMixin, UsesStoredWorkflowMixin, UsesVisualizationMixin, UsesItemRatings):
    _page_list = PageListGrid()
    _all_published_list = PageAllPublishedGrid()
    _history_selection_grid = HistorySelectionGrid()
    _workflow_selection_grid = WorkflowSelectionGrid()
    _datasets_selection_grid = HistoryDatasetAssociationSelectionGrid()
    _page_selection_grid = PageSelectionGrid()
    _visualization_selection_grid = VisualizationSelectionGrid()
    page_manager: PageManager = depends(PageManager)
    history_manager: HistoryManager = depends(HistoryManager)
    history_serializer: HistorySerializer = depends(HistorySerializer)
    hda_manager: HDAManager = depends(HDAManager)
    workflow_manager: WorkflowsManager = depends(WorkflowsManager)
    slug_builder: SlugBuilder = depends(SlugBuilder)

    def __init__(self, app: StructuredApp):
        super().__init__(app)

    @web.expose
    @web.json
    @web.require_login()
    def list(self, trans, *args, **kwargs):
        """List user's pages."""
        # Handle operation
        if "operation" in kwargs and "id" in kwargs:
            session = trans.sa_session
            operation = kwargs["operation"].lower()
            ids = util.listify(kwargs["id"])
            for id in ids:
                if operation == "delete":
                    item = session.query(model.Page).get(self.decode_id(id))
                    self.security_check(trans, item, check_ownership=True)
                    item.deleted = True
            with transaction(session):
                session.commit()

        # Build grid dictionary.
        grid = self._page_list(trans, *args, **kwargs)
        grid["shared_by_others"] = self._get_shared(trans)
        return grid

    @web.expose
    @web.json
    def list_published(self, trans, *args, **kwargs):
        grid = self._all_published_list(trans, *args, **kwargs)
        grid["shared_by_others"] = self._get_shared(trans)
        return grid

    def _get_shared(self, trans):
        """Identify shared pages"""
        shared_by_others = (
            trans.sa_session.query(model.PageUserShareAssociation)
            .filter_by(user=trans.get_user())
            .join(model.Page.table)
            .filter(model.Page.deleted == false())
            .order_by(desc(model.Page.update_time))
            .all()
        )
        return [
            {"username": p.page.user.username, "slug": p.page.slug, "title": p.page.title} for p in shared_by_others
        ]

    @web.expose_api
    @web.require_login("create pages")
    def create(self, trans, payload=None, **kwd):
        """
        Create a new page.
        """
        if trans.request.method == "GET":
            form_title = "Create new Page"
            title = ""
            slug = ""
            content = ""
            content_hide = True
            if "invocation_id" in kwd:
                invocation_id = kwd.get("invocation_id")
                form_title = f"{form_title} from Invocation Report"
                slug = f"invocation-report-{invocation_id}"
                invocation_report = self.workflow_manager.get_invocation_report(
                    trans, trans.security.decode_id(invocation_id)
                )
                title = invocation_report.get("title")
                content = invocation_report.get("markdown")
                content_hide = False
            return {
                "title": form_title,
                "inputs": [
                    {
                        "name": "title",
                        "label": "Name",
                        "value": title,
                    },
                    {
                        "name": "slug",
                        "label": "Identifier",
                        "help": "A unique identifier that will be used for public links to this page. This field can only contain lowercase letters, numbers, and dashes (-).",
                        "value": slug,
                    },
                    {
                        "name": "annotation",
                        "label": "Annotation",
                        "help": "A description of the page. The annotation is shown alongside published pages.",
                    },
                    {
                        "name": "content_format",
                        "label": "Content Format",
                        "value": "markdown",
                        "hidden": True,
                    },
                    {
                        "name": "content",
                        "label": "Content",
                        "area": True,
                        "value": content,
                        "hidden": content_hide,
                    },
                ],
            }
        else:
            page = self.page_manager.create_page(trans, CreatePagePayload(**payload))
            return {"message": "Page '%s' successfully created." % page.title, "status": "success"}

    @web.legacy_expose_api
    @web.require_login("edit pages")
    def edit(self, trans, payload=None, **kwd):
        """
        Edit a page's attributes.
        """
        id = kwd.get("id")
        if not id:
            return self.message_exception(trans, "No page id received for editing.")
        decoded_id = self.decode_id(id)
        user = trans.get_user()
        p = trans.sa_session.query(model.Page).get(decoded_id)
        p = self.security_check(trans, p, check_ownership=True)
        if trans.request.method == "GET":
            if p.slug is None:
                self.slug_builder.create_item_slug(trans.sa_session, p)
            return {
                "title": "Edit page attributes",
                "inputs": [
                    {"name": "title", "label": "Name", "value": p.title},
                    {
                        "name": "slug",
                        "label": "Identifier",
                        "value": p.slug,
                        "help": "A unique identifier that will be used for public links to this page. This field can only contain lowercase letters, numbers, and dashes (-).",
                    },
                    {
                        "name": "annotation",
                        "label": "Annotation",
                        "value": self.get_item_annotation_str(trans.sa_session, user, p),
                        "help": "A description of the page. The annotation is shown alongside published pages.",
                    },
                ],
            }
        else:
            p_title = payload.get("title")
            p_slug = payload.get("slug")
            p_annotation = payload.get("annotation")
            if not p_title:
                return self.message_exception(trans, "Please provide a page name is required.")
            elif not p_slug:
                return self.message_exception(trans, "Please provide a unique identifier.")
            elif not self._is_valid_slug(p_slug):
                return self.message_exception(
                    trans, "Page identifier can only contain lowercase letters, numbers, and dashes (-)."
                )
            elif (
                p_slug != p.slug
                and trans.sa_session.query(model.Page).filter_by(user=p.user, slug=p_slug, deleted=False).first()
            ):
                return self.message_exception(trans, "Page id must be unique.")
            else:
                p.title = p_title
                p.slug = p_slug
                if p_annotation:
                    p_annotation = sanitize_html(p_annotation)
                    self.add_item_annotation(trans.sa_session, user, p, p_annotation)
                trans.sa_session.add(p)
                with transaction(trans.sa_session):
                    trans.sa_session.commit()
            return {"message": "Attributes of '%s' successfully saved." % p.title, "status": "success"}

    @web.expose
    @web.require_login()
    def display(self, trans, id, **kwargs):
        id = self.decode_id(id)
        page = trans.sa_session.query(model.Page).get(id)
        if not page:
            raise web.httpexceptions.HTTPNotFound()
        return self.display_by_username_and_slug(trans, page.user.username, page.slug)

    @web.expose
    def display_by_username_and_slug(self, trans, username, slug, **kwargs):
        """Display page based on a username and slug."""

        # Get page.
        session = trans.sa_session
        user = session.query(model.User).filter_by(username=username).first()
        page = trans.sa_session.query(model.Page).filter_by(user=user, slug=slug, deleted=False).first()
        if page is None:
            raise web.httpexceptions.HTTPNotFound()

        # Security check raises error if user cannot access page.
        self.security_check(trans, page, False, True)

        # Encode page identifier.
        page_id = trans.security.encode_id(page.id)

        # Redirect to client.
        return trans.response.send_redirect(
            web.url_for(
                controller="published",
                action="page",
                id=page_id,
            )
        )

    @web.expose
    @web.json
    @web.require_login("select a history from saved histories")
    def list_histories_for_selection(self, trans, **kwargs):
        """Returns HTML that enables a user to select one or more histories."""
        return self._history_selection_grid(trans, **kwargs)

    @web.expose
    @web.json
    @web.require_login("select a workflow from saved workflows")
    def list_workflows_for_selection(self, trans, **kwargs):
        """Returns HTML that enables a user to select one or more workflows."""
        return self._workflow_selection_grid(trans, **kwargs)

    @web.expose
    @web.json
    @web.require_login("select a visualization from saved visualizations")
    def list_visualizations_for_selection(self, trans, **kwargs):
        """Returns HTML that enables a user to select one or more visualizations."""
        return self._visualization_selection_grid(trans, **kwargs)

    @web.expose
    @web.json
    @web.require_login("select a page from saved pages")
    def list_pages_for_selection(self, trans, **kwargs):
        """Returns HTML that enables a user to select one or more pages."""
        return self._page_selection_grid(trans, **kwargs)

    @web.expose
    @web.json
    @web.require_login("select a dataset from saved datasets")
    def list_datasets_for_selection(self, trans, **kwargs):
        """Returns HTML that enables a user to select one or more datasets."""
        return self._datasets_selection_grid(trans, **kwargs)

    def get_page(self, trans, id, check_ownership=True, check_accessible=False):
        """Get a page from the database by id."""
        # Load history from database
        id = self.decode_id(id)
        page = trans.sa_session.query(model.Page).get(id)
        if not page:
            error("Page not found")
        else:
            return self.security_check(trans, page, check_ownership, check_accessible)

    def get_item(self, trans, id):
        return self.get_page(trans, id)



import abc
from typing import Optional

from sqlalchemy.orm import scoped_session

from galaxy.security.idencoding import IdEncodingHelper
from galaxy.work.context import (
    GalaxyAbstractRequest,
    GalaxyAbstractResponse,
)
from tool_shed.structured_app import ToolShedApp
from tool_shed.webapp.model import (
    GalaxySession,
    User,
)
from tool_shed.webapp.model.mapping import ToolShedModelMapping


class ProvidesAppContext:
    """For transaction-like objects to provide the shed convenience layer for
    database and event handling.

    Mixed in class must provide `app` property.
    """

    @abc.abstractproperty
    def app(self) -> ToolShedApp:
        """Provide access to the shed ``app`` object."""

    @property
    def security(self) -> IdEncodingHelper:
        return self.app.security

    @property
    def sa_session(self) -> scoped_session:
        """Provide access to Galaxy's SQLAlchemy session.

        :rtype: galaxy.model.scoped_session.galaxy_scoped_session
        """
        return self.model.session

    @property
    def model(self) -> ToolShedModelMapping:
        """Provide access to Tool Shed's model mapping class."""
        return self.app.model


class ProvidesUserContext(ProvidesAppContext):
    """For transaction-like objects to provide Galaxy convenience layer for
    reasoning about users.

    Mixed in class must provide `user` and `app`
    properties.
    """

    @abc.abstractproperty
    def user(self) -> Optional[User]:
        """Provide access to the user object."""

    @property
    def anonymous(self) -> bool:
        return self.user is None

    @property
    def user_is_admin(self) -> bool:
        return self.app.config.is_admin_user(self.user)

    @property
    def user_is_bootstrap_admin(self) -> bool:
        """Master key provided so there is no real user"""
        user = self.user
        return not self.anonymous and user is not None and user.bootstrap_admin_user


class SessionRequestContext(ProvidesUserContext):
    @abc.abstractmethod
    def get_galaxy_session(self) -> Optional[GalaxySession]:
        ...

    @abc.abstractproperty
    def request(self) -> GalaxyAbstractRequest:
        ...

    @abc.abstractproperty
    def response(self) -> GalaxyAbstractResponse:
        ...



"""
API Controller providing Galaxy Tours
"""
import logging

from galaxy.tours import (
    TourDetails,
    TourList,
    ToursRegistry,
)
from . import (
    depends,
    Router,
)

log = logging.getLogger(__name__)


router = Router(tags=["tours"])


@router.cbv
class FastAPITours:
    registry: ToursRegistry = depends(ToursRegistry)  # type: ignore[type-abstract]  # https://github.com/python/mypy/issues/4717

    @router.get("/api/tours")
    def index(self) -> TourList:
        """Return list of available tours."""
        return self.registry.get_tours()

    @router.get("/api/tours/{tour_id}")
    def show(self, tour_id: str) -> TourDetails:
        """Return a tour definition."""
        return self.registry.tour_contents(tour_id)

    @router.post("/api/tours/{tour_id}", require_admin=True)
    def update_tour(self, tour_id: str) -> TourDetails:
        """Return a tour definition."""
        return self.registry.load_tour(tour_id)



import logging
from typing import (
    Any,
    Callable,
    Dict,
)

import tool_shed.util.shed_util_common as suc
from galaxy import (
    exceptions,
    util,
    web,
)
from galaxy.model.base import transaction
from galaxy.web import (
    expose_api,
    expose_api_anonymous_and_sessionless,
    require_admin,
)
from galaxy.webapps.base.controller import BaseAPIController
from tool_shed.util import repository_util

log = logging.getLogger(__name__)


class CategoriesController(BaseAPIController):
    """RESTful controller for interactions with categories in the Tool Shed."""

    def __get_value_mapper(self, trans) -> Dict[str, Callable]:
        value_mapper = {"id": trans.security.encode_id}
        return value_mapper

    @expose_api
    @require_admin
    def create(self, trans, payload, **kwd):
        """
        POST /api/categories
        Return a dictionary of information about the created category.
        The following parameters are included in the payload:

        :param name (required): the name of the category
        :param description (optional): the description of the category (if not provided, the name will be used)

        Example: POST /api/categories/?key=XXXYYYXXXYYY
        Content-Disposition: form-data; name="name" Category_Name
        Content-Disposition: form-data; name="description" Category_Description
        """
        category_dict = dict(message="", status="ok")
        name = payload.get("name", "")
        if name:
            description = payload.get("description", "")
            if not description:
                # Default the description to the name.
                description = name
            if suc.get_category_by_name(self.app, name):
                raise exceptions.Conflict("A category with that name already exists.")
            else:
                # Create the category
                category = self.app.model.Category(name=name, description=description)
                trans.sa_session.add(category)
                with transaction(trans.sa_session):
                    trans.sa_session.commit()
                category_dict = category.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
                category_dict["message"] = f"Category '{str(category.name)}' has been created"
                category_dict["url"] = web.url_for(
                    controller="categories", action="show", id=trans.security.encode_id(category.id)
                )
        else:
            raise exceptions.RequestParameterMissingException('Missing required parameter "name".')
        return category_dict

    @expose_api_anonymous_and_sessionless
    def get_repositories(self, trans, category_id, **kwd):
        """
        GET /api/categories/{encoded_category_id}/repositories
        Return information about the provided category and the repositories in that category.

        :param id: the encoded id of the Category object
        :param sort_key: the field by which the repositories should be sorted
        :param sort_order: ascending or descending sort
        :param page: the page number to return

        Example: GET localhost:9009/api/categories/f9cad7b01a472135/repositories
        """
        installable = util.asbool(kwd.get("installable", "false"))
        sort_key = kwd.get("sort_key", "name")
        sort_order = kwd.get("sort_order", "asc")
        page = kwd.get("page", None)
        category = suc.get_category(self.app, category_id)
        category_dict: Dict[str, Any]
        if category is None:
            category_dict = dict(message=f"Unable to locate category record for id {str(id)}.", status="error")
            return category_dict
        category_dict = category.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        category_dict["repository_count"] = suc.count_repositories_in_category(self.app, category_id)
        category_dict["url"] = web.url_for(
            controller="categories", action="show", id=trans.security.encode_id(category.id)
        )
        repositories = repository_util.get_repositories_by_category(
            self.app, category.id, installable=installable, sort_order=sort_order, sort_key=sort_key, page=page
        )
        category_dict["repositories"] = repositories
        return category_dict

    @expose_api_anonymous_and_sessionless
    def index(self, trans, deleted=False, **kwd):
        """
        GET /api/categories
        Return a list of dictionaries that contain information about each Category.

        :param deleted: flag used to include deleted categories

        Example: GET localhost:9009/api/categories
        """
        category_dicts = []
        deleted = util.asbool(deleted)
        if deleted and not trans.user_is_admin:
            raise exceptions.AdminRequiredException("Only administrators can query deleted categories.")
        for category in (
            trans.sa_session.query(self.app.model.Category)
            .filter(self.app.model.Category.table.c.deleted == deleted)
            .order_by(self.app.model.Category.table.c.name)
        ):
            category_dict = category.to_dict(view="collection", value_mapper=self.__get_value_mapper(trans))
            category_dict["url"] = web.url_for(
                controller="categories", action="show", id=trans.security.encode_id(category.id)
            )
            category_dict[
                "repositories"
            ] = self.app.repository_registry.viewable_repositories_and_suites_by_category.get(category.name, 0)
            category_dicts.append(category_dict)
        return category_dicts

    @expose_api_anonymous_and_sessionless
    def show(self, trans, id, **kwd):
        """
        GET /api/categories/{encoded_category_id}
        Return a dictionary of information about a category.

        :param id: the encoded id of the Category object

        Example: GET localhost:9009/api/categories/f9cad7b01a472135
        """
        category = suc.get_category(self.app, id)
        if category is None:
            category_dict = dict(message=f"Unable to locate category record for id {str(id)}.", status="error")
            return category_dict
        category_dict = category.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        category_dict["url"] = web.url_for(
            controller="categories", action="show", id=trans.security.encode_id(category.id)
        )
        return category_dict



"""
API operations allowing clients to determine datatype supported by Galaxy.
"""
import logging
from typing import (
    cast,
    Dict,
    List,
    Optional,
    Union,
)

from fastapi import Query

from galaxy.datatypes.registry import Registry
from galaxy.managers.datatypes import (
    DatatypeConverterList,
    DatatypeDetails,
    DatatypesCombinedMap,
    DatatypesEDAMDetailsDict,
    DatatypesMap,
    view_converters,
    view_edam_data,
    view_edam_formats,
    view_index,
    view_mapping,
    view_sniffers,
)
from . import (
    depends,
    Router,
)

log = logging.getLogger(__name__)

router = Router(tags=["datatypes"])

ExtensionOnlyQueryParam: Optional[bool] = Query(
    default=True,
    title="Extension only",
    description="Whether to return only the datatype's extension rather than the datatype's details",
)

UploadOnlyQueryParam: Optional[bool] = Query(
    default=True,
    title="Upload only",
    description="Whether to return only datatypes which can be uploaded",
)

IdentifierOnly: Optional[bool] = Query(
    default=True,
    title="prefixIRI only",
    description="Whether to return only the EDAM prefixIRI rather than the EDAM details",
)


@router.cbv
class FastAPIDatatypes:
    datatypes_registry: Registry = depends(Registry)

    @router.get(
        "/api/datatypes",
        summary="Lists all available data types",
        response_description="List of data types",
    )
    async def index(
        self,
        extension_only: Optional[bool] = ExtensionOnlyQueryParam,
        upload_only: Optional[bool] = UploadOnlyQueryParam,
    ) -> Union[List[DatatypeDetails], List[str]]:
        """Gets the list of all available data types."""
        return view_index(self.datatypes_registry, extension_only, upload_only)

    @router.get(
        "/api/datatypes/mapping",
        summary="Returns mappings for data types and their implementing classes",
        response_description="Dictionary to map data types with their classes",
    )
    async def mapping(self) -> DatatypesMap:
        """Gets mappings for data types."""
        return view_mapping(self.datatypes_registry)

    @router.get(
        "/api/datatypes/types_and_mapping",
        summary="Returns all the data types extensions and their mappings",
        response_description="Dictionary to map data types with their classes",
    )
    async def types_and_mapping(
        self,
        extension_only: Optional[bool] = ExtensionOnlyQueryParam,
        upload_only: Optional[bool] = UploadOnlyQueryParam,
    ) -> DatatypesCombinedMap:
        """Combines the datatype information from (/api/datatypes) and the
        mapping information from (/api/datatypes/mapping) into a single
        response."""
        return DatatypesCombinedMap(
            datatypes=view_index(self.datatypes_registry, extension_only, upload_only),
            datatypes_mapping=view_mapping(self.datatypes_registry),
        )

    @router.get(
        "/api/datatypes/sniffers",
        summary="Returns the list of all installed sniffers",
        response_description="List of datatype sniffers",
    )
    async def sniffers(self) -> List[str]:
        """Gets the list of all installed data type sniffers."""
        return view_sniffers(self.datatypes_registry)

    @router.get(
        "/api/datatypes/converters",
        summary="Returns the list of all installed converters",
        response_description="List of all datatype converters",
    )
    async def converters(self) -> DatatypeConverterList:
        """Gets the list of all installed converters."""
        return view_converters(self.datatypes_registry)

    @router.get(
        "/api/datatypes/edam_formats",
        summary="Returns a dictionary/map of datatypes and EDAM formats",
        response_description="Dictionary/map of datatypes and EDAM formats",
    )
    async def edam_formats(self) -> Dict[str, str]:
        """Gets a map of datatypes and their corresponding EDAM formats."""
        return cast(Dict[str, str], view_edam_formats(self.datatypes_registry))

    @router.get(
        "/api/datatypes/edam_formats/detailed",
        summary="Returns a dictionary of datatypes and EDAM format details",
        response_description="Dictionary of EDAM format details containing the EDAM iri, label, and definition",
        response_model=DatatypesEDAMDetailsDict,
    )
    async def edam_formats_detailed(self):
        """Gets a map of datatypes and their corresponding EDAM formats.
        EDAM formats contain the EDAM iri, label, and definition."""
        return view_edam_formats(self.datatypes_registry, True)

    @router.get(
        "/api/datatypes/edam_data",
        summary="Returns a dictionary/map of datatypes and EDAM data",
        response_description="Dictionary/map of datatypes and EDAM data",
    )
    async def edam_data(self) -> Dict[str, str]:
        """Gets a map of datatypes and their corresponding EDAM data."""
        return cast(Dict[str, str], view_edam_data(self.datatypes_registry))

    @router.get(
        "/api/datatypes/edam_data/detailed",
        summary="Returns a dictionary of datatypes and EDAM data details",
        response_description="Dictionary of EDAM data details containing the EDAM iri, label, and definition",
        response_model=DatatypesEDAMDetailsDict,
    )
    async def edam_data_detailed(self):
        """Gets a map of datatypes and their corresponding EDAM data.
        EDAM data contains the EDAM iri, label, and definition."""
        return view_edam_data(self.datatypes_registry, True)



"""
Visualizations resource control over the API.

NOTE!: this is a work in progress and functionality and data structures
may change often.
"""
import json
import logging

from fastapi import (
    Body,
    Path,
    Response,
    status,
)

from galaxy import (
    exceptions,
    util,
    web,
)
from galaxy.managers.context import ProvidesUserContext
from galaxy.model.base import transaction
from galaxy.model.item_attrs import UsesAnnotations
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.schema import (
    SetSlugPayload,
    ShareWithPayload,
    ShareWithStatus,
    SharingStatus,
)
from galaxy.util.hash_util import md5_hash_str
from galaxy.web import expose_api
from galaxy.webapps.base.controller import UsesVisualizationMixin
from galaxy.webapps.base.webapp import GalaxyWebTransaction
from galaxy.webapps.galaxy.api import (
    BaseGalaxyAPIController,
    depends,
    DependsOnTrans,
    Router,
)
from galaxy.webapps.galaxy.services.visualizations import VisualizationsService

log = logging.getLogger(__name__)

router = Router(tags=["visualizations"])

VisualizationIdPathParam: DecodedDatabaseIdField = Path(
    ..., title="Visualization ID", description="The encoded database identifier of the Visualization."
)


@router.cbv
class FastAPIVisualizations:
    service: VisualizationsService = depends(VisualizationsService)

    @router.get(
        "/api/visualizations/{id}/sharing",
        summary="Get the current sharing status of the given Page.",
    )
    def sharing(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
    ) -> SharingStatus:
        """Return the sharing status of the item."""
        return self.service.shareable_service.sharing(trans, id)

    @router.put(
        "/api/visualizations/{id}/enable_link_access",
        summary="Makes this item accessible by a URL link.",
    )
    def enable_link_access(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
    ) -> SharingStatus:
        """Makes this item accessible by a URL link and return the current sharing status."""
        return self.service.shareable_service.enable_link_access(trans, id)

    @router.put(
        "/api/visualizations/{id}/disable_link_access",
        summary="Makes this item inaccessible by a URL link.",
    )
    def disable_link_access(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
    ) -> SharingStatus:
        """Makes this item inaccessible by a URL link and return the current sharing status."""
        return self.service.shareable_service.disable_link_access(trans, id)

    @router.put(
        "/api/visualizations/{id}/publish",
        summary="Makes this item public and accessible by a URL link.",
    )
    def publish(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
    ) -> SharingStatus:
        """Makes this item publicly available by a URL link and return the current sharing status."""
        return self.service.shareable_service.publish(trans, id)

    @router.put(
        "/api/visualizations/{id}/unpublish",
        summary="Removes this item from the published list.",
    )
    def unpublish(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
    ) -> SharingStatus:
        """Removes this item from the published list and return the current sharing status."""
        return self.service.shareable_service.unpublish(trans, id)

    @router.put(
        "/api/visualizations/{id}/share_with_users",
        summary="Share this item with specific users.",
    )
    def share_with_users(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
        payload: ShareWithPayload = Body(...),
    ) -> ShareWithStatus:
        """Shares this item with specific users and return the current sharing status."""
        return self.service.shareable_service.share_with_users(trans, id, payload)

    @router.put(
        "/api/visualizations/{id}/slug",
        summary="Set a new slug for this shared item.",
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def set_slug(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        id: DecodedDatabaseIdField = VisualizationIdPathParam,
        payload: SetSlugPayload = Body(...),
    ):
        """Sets a new slug to access this item by URL. The new slug must be unique."""
        self.service.shareable_service.set_slug(trans, id, payload)
        return Response(status_code=status.HTTP_204_NO_CONTENT)


class VisualizationsController(BaseGalaxyAPIController, UsesVisualizationMixin, UsesAnnotations):
    """
    RESTful controller for interactions with visualizations.
    """

    service: VisualizationsService = depends(VisualizationsService)

    @expose_api
    def index(self, trans: GalaxyWebTransaction, **kwargs):
        """
        GET /api/visualizations:
        """
        rval = []
        user = trans.user

        # TODO: search for: title, made by user, creation time range, type (vis name), dbkey, etc.
        # TODO: limit, offset, order_by
        # TODO: deleted

        # this is the default search - user's vis, vis shared with user, published vis
        visualizations = self.get_visualizations_by_user(trans, user)
        visualizations += self.get_visualizations_shared_with_user(trans, user)
        visualizations += self.get_published_visualizations(trans, exclude_user=user)
        # TODO: the admin case - everything

        for visualization in visualizations:
            item = self.get_visualization_summary_dict(visualization)
            item = trans.security.encode_dict_ids(item)
            item["url"] = web.url_for("visualization", id=item["id"])
            rval.append(item)

        return rval

    @expose_api
    def show(self, trans: GalaxyWebTransaction, id: str, **kwargs):
        """
        GET /api/visualizations/{viz_id}
        """
        # TODO: revisions should be a contents/nested controller like viz/xxx/r/xxx)?
        # the important thing is the config
        # TODO:?? /api/visualizations/registry -> json of registry.listings?
        visualization = self.get_visualization(trans, id, check_ownership=False, check_accessible=True)
        dictionary = trans.security.encode_dict_ids(self.get_visualization_dict(visualization))
        dictionary["url"] = web.url_for(
            controller="visualization",
            action="display_by_username_and_slug",
            username=visualization.user.username,
            slug=visualization.slug,
        )
        dictionary["username"] = visualization.user.username
        dictionary["email_hash"] = md5_hash_str(visualization.user.email)
        dictionary["tags"] = visualization.make_tag_string_list()
        dictionary["annotation"] = self.get_item_annotation_str(trans.sa_session, trans.user, visualization)
        # need to encode ids in revisions as well
        encoded_revisions = []
        for revision in dictionary["revisions"]:
            # NOTE: does not encode ids inside the configs
            encoded_revisions.append(trans.security.encode_id(revision))
        dictionary["revisions"] = encoded_revisions
        dictionary["latest_revision"] = trans.security.encode_dict_ids(dictionary["latest_revision"])
        if trans.app.visualizations_registry:
            visualization = trans.app.visualizations_registry.get_plugin(dictionary["type"])
            dictionary["plugin"] = visualization.to_dict()
        return dictionary

    @expose_api
    def create(self, trans: GalaxyWebTransaction, payload: dict, **kwargs):
        """
        POST /api/visualizations
        creates a new visualization using the given payload

        POST /api/visualizations?import_id={encoded_visualization_id}
        imports a copy of an existing visualization into the user's workspace
        """
        rval = None

        if "import_id" in payload:
            import_id = payload["import_id"]
            visualization = self.import_visualization(trans, import_id, user=trans.user)

        else:
            payload = self._validate_and_parse_payload(payload)
            # must have a type (I've taken this to be the visualization name)
            if "type" not in payload:
                raise exceptions.RequestParameterMissingException("key/value 'type' is required")
            vis_type = payload.pop("type", False)

            payload["save"] = True
            try:
                # generate defaults - this will err if given a weird key?
                visualization = self.create_visualization(trans, vis_type, **payload)
            except ValueError as val_err:
                raise exceptions.RequestParameterMissingException(str(val_err))

        rval = {"id": trans.security.encode_id(visualization.id)}

        return rval

    @expose_api
    def update(self, trans: GalaxyWebTransaction, id: str, payload: dict, **kwargs):
        """
        PUT /api/visualizations/{encoded_visualization_id}
        """
        rval = None

        payload = self._validate_and_parse_payload(payload)

        # there's a differentiation here between updating the visualiztion and creating a new revision
        #   that needs to be handled clearly here
        # or alternately, using a different controller like PUT /api/visualizations/{id}/r/{id}

        # TODO: consider allowing direct alteration of revisions title (without a new revision)
        #   only create a new revsion on a different config

        # only update owned visualizations
        visualization = self.get_visualization(trans, id, check_ownership=True)
        title = payload.get("title", visualization.latest_revision.title)
        dbkey = payload.get("dbkey", visualization.latest_revision.dbkey)
        config = payload.get("config", visualization.latest_revision.config)

        latest_config = visualization.latest_revision.config
        if (
            (title != visualization.latest_revision.title)
            or (dbkey != visualization.latest_revision.dbkey)
            or (json.dumps(config) != json.dumps(latest_config))
        ):
            revision = self.add_visualization_revision(trans, visualization, config, title, dbkey)
            rval = {"id": id, "revision": revision.id}

        # allow updating vis title
        visualization.title = title
        with transaction(trans.sa_session):
            trans.sa_session.commit()

        return rval

    def _validate_and_parse_payload(self, payload):
        """
        Validate and parse incomming data payload for a visualization.
        """
        # This layer handles (most of the stricter idiot proofing):
        #   - unknown/unallowed keys
        #   - changing data keys from api key to attribute name
        #   - protection against bad data form/type
        #   - protection against malicious data content
        # all other conversions and processing (such as permissions, etc.) should happen down the line

        # keys listed here don't error when attempting to set, but fail silently
        #   this allows PUT'ing an entire model back to the server without attribute errors on uneditable attrs
        valid_but_uneditable_keys = (
            "id",
            "model_class"
            # TODO: fill out when we create to_dict, get_dict, whatevs
        )
        # TODO: deleted
        # TODO: importable
        ValidationError = exceptions.RequestParameterInvalidException

        validated_payload = {}
        for key, val in payload.items():
            # TODO: validate types in VALID_TYPES/registry names at the mixin/model level?
            if key == "type":
                if not isinstance(val, str):
                    raise ValidationError(f"{key} must be a string or unicode: {str(type(val))}")
                val = util.sanitize_html.sanitize_html(val)
            elif key == "config":
                if not isinstance(val, dict):
                    raise ValidationError(f"{key} must be a dictionary: {str(type(val))}")

            elif key == "annotation":
                if not isinstance(val, str):
                    raise ValidationError(f"{key} must be a string or unicode: {str(type(val))}")
                val = util.sanitize_html.sanitize_html(val)

            # these are keys that actually only be *updated* at the revision level and not here
            #   (they are still valid for create, tho)
            elif key == "title":
                if not isinstance(val, str):
                    raise ValidationError(f"{key} must be a string or unicode: {str(type(val))}")
                val = util.sanitize_html.sanitize_html(val)
            elif key == "slug":
                if not isinstance(val, str):
                    raise ValidationError(f"{key} must be a string: {str(type(val))}")
                val = util.sanitize_html.sanitize_html(val)
            elif key == "dbkey":
                if not isinstance(val, str):
                    raise ValidationError(f"{key} must be a string or unicode: {str(type(val))}")
                val = util.sanitize_html.sanitize_html(val)

            elif key not in valid_but_uneditable_keys:
                continue
                # raise AttributeError( 'unknown key: %s' %( str( key ) ) )

            validated_payload[key] = val
        return validated_payload



"""
API operations on Galaxy's object store.
"""
import logging
from typing import List

from fastapi import (
    Path,
    Query,
)

from galaxy.exceptions import (
    ObjectNotFound,
    RequestParameterInvalidException,
)
from galaxy.managers.context import ProvidesUserContext
from galaxy.objectstore import (
    BaseObjectStore,
    ConcreteObjectStoreModel,
)
from . import (
    depends,
    DependsOnTrans,
    Router,
)

log = logging.getLogger(__name__)

router = Router(tags=["object sstore"])

ConcreteObjectStoreIdPathParam: str = Path(
    ..., title="Concrete Object Store ID", description="The concrete object store ID."
)

SelectableQueryParam: bool = Query(
    default=False,
    title="Selectable",
    description="Restrict index query to user selectable object stores, the current implementation requires this to be true.",
)


@router.cbv
class FastAPIObjectStore:
    object_store: BaseObjectStore = depends(BaseObjectStore)

    @router.get(
        "/api/object_stores",
        summary="Get a list of (currently only concrete) object stores configured with this Galaxy instance.",
        response_description="A list of the configured object stores.",
    )
    def index(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        selectable: bool = SelectableQueryParam,
    ) -> List[ConcreteObjectStoreModel]:
        if not selectable:
            raise RequestParameterInvalidException(
                "The object store index query currently needs to be called with selectable=true"
            )
        selectable_ids = self.object_store.object_store_ids_allowing_selection()
        return [self._model_for(selectable_id) for selectable_id in selectable_ids]

    @router.get(
        "/api/object_stores/{object_store_id}",
        summary="Get information about a concrete object store configured with Galaxy.",
    )
    def show_info(
        self,
        trans: ProvidesUserContext = DependsOnTrans,
        object_store_id: str = ConcreteObjectStoreIdPathParam,
    ) -> ConcreteObjectStoreModel:
        return self._model_for(object_store_id)

    def _model_for(self, object_store_id: str) -> ConcreteObjectStoreModel:
        concrete_object_store = self.object_store.get_concrete_store_by_object_store_id(object_store_id)
        if concrete_object_store is None:
            raise ObjectNotFound()
        return concrete_object_store.to_model(object_store_id)



"""
This module *does not* contain API routes. It exclusively contains dependencies to be used in FastAPI routes
"""
import inspect
from enum import Enum
from string import Template
from typing import (
    Any,
    AsyncGenerator,
    cast,
    MutableMapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
)
from urllib.parse import (
    urlencode,
    urljoin,
)

from a2wsgi.wsgi import build_environ
from fastapi import (
    Form,
    Header,
    Query,
    Request,
    Response,
    Security,
)
from fastapi.exceptions import RequestValidationError
from fastapi.params import Depends
from fastapi.routing import APIRoute
from fastapi.security import (
    APIKeyCookie,
    APIKeyHeader,
    APIKeyQuery,
)
from fastapi_utils.cbv import cbv
from fastapi_utils.inferring_router import InferringRouter
from pydantic import ValidationError
from pydantic.main import BaseModel
from starlette.datastructures import Headers
from starlette.routing import (
    Match,
    NoMatchFound,
)
from starlette.types import Scope

try:
    from starlette_context import context as request_context
except ImportError:
    request_context = None  # type: ignore[assignment]

from galaxy import (
    app as galaxy_app,
    model,
    web,
)
from galaxy.exceptions import (
    AdminRequiredException,
    UserCannotRunAsException,
)
from galaxy.managers.session import GalaxySessionManager
from galaxy.managers.users import UserManager
from galaxy.model import User
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.structured_app import StructuredApp
from galaxy.web.framework.decorators import require_admin_message
from galaxy.webapps.base.controller import BaseAPIController
from galaxy.work.context import (
    GalaxyAbstractRequest,
    GalaxyAbstractResponse,
    SessionRequestContext,
)

api_key_query = APIKeyQuery(name="key", auto_error=False)
api_key_header = APIKeyHeader(name="x-api-key", auto_error=False)
api_key_cookie = APIKeyCookie(name="galaxysession", auto_error=False)


def get_app() -> StructuredApp:
    return cast(StructuredApp, galaxy_app.app)


async def get_app_with_request_session() -> AsyncGenerator[StructuredApp, None]:
    app = get_app()
    request_id = request_context.data["X-Request-ID"]
    app.model.set_request_id(request_id)
    try:
        yield app
    finally:
        app.model.unset_request_id(request_id)


DependsOnApp = cast(StructuredApp, Depends(get_app_with_request_session))


T = TypeVar("T")


class GalaxyTypeDepends(Depends):
    """Variant of fastapi Depends that can also work on WSGI Galaxy controllers."""

    def __init__(self, callable, dep_type):
        super().__init__(callable)
        self.galaxy_type_depends = dep_type


def depends(dep_type: Type[T]) -> T:
    def _do_resolve(request: Request):
        return get_app().resolve(dep_type)

    return cast(T, GalaxyTypeDepends(_do_resolve, dep_type))


def get_session_manager(app: StructuredApp = DependsOnApp) -> GalaxySessionManager:
    # TODO: find out how to adapt dependency for Galaxy/Report/TS
    return GalaxySessionManager(app.model)


def get_session(
    session_manager=cast(GalaxySessionManager, Depends(get_session_manager)),
    security: IdEncodingHelper = depends(IdEncodingHelper),
    galaxysession: str = Security(api_key_cookie),
) -> Optional[model.GalaxySession]:
    if galaxysession:
        session_key = security.decode_guid(galaxysession)
        if session_key:
            return session_manager.get_session_from_session_key(session_key)
        # TODO: What should we do if there is no session? Since this is the API, maybe nothing is the right choice?
    return None


def get_api_user(
    user_manager: UserManager = depends(UserManager),
    key: str = Security(api_key_query),
    x_api_key: str = Security(api_key_header),
    run_as: Optional[DecodedDatabaseIdField] = Header(
        default=None,
        title="Run as User",
        description=(
            "The user ID that will be used to effectively make this API call. "
            "Only admins and designated users can make API calls on behalf of other users."
        ),
    ),
) -> Optional[User]:
    api_key = key or x_api_key
    if not api_key:
        return None
    user = user_manager.by_api_key(api_key=api_key)
    if run_as:
        if user_manager.user_can_do_run_as(user):
            return user_manager.by_id(run_as)
        else:
            raise UserCannotRunAsException
    return user


def get_user(
    galaxy_session=cast(Optional[model.GalaxySession], Depends(get_session)),
    api_user=cast(Optional[User], Depends(get_api_user)),
) -> Optional[User]:
    if galaxy_session:
        return galaxy_session.user
    return api_user


class UrlBuilder:
    def __init__(self, request: Request):
        self.request = request

    def __call__(self, name: str, **path_params):
        qualified = path_params.pop("qualified", False)
        # starlette does not support query parameters in url_path_for: https://github.com/encode/starlette/issues/560
        query_params = path_params.pop("query_params", None)
        try:
            if qualified:
                url = str(self.request.url_for(name, **path_params))
            else:
                url = self.request.app.url_path_for(name, **path_params)
            if query_params:
                url = f"{url}?{urlencode(query_params)}"
            return url
        except NoMatchFound:
            # Fallback to legacy url_for
            if query_params:
                path_params.update(query_params)
            return web.url_for(name, **path_params)


class GalaxyASGIRequest(GalaxyAbstractRequest):
    """Wrapper around Starlette/FastAPI Request object.

    Implements the GalaxyAbstractRequest interface to provide access to some properties
    of the request commonly used."""

    def __init__(self, request: Request):
        self.__request = request
        self.__environ: Optional[MutableMapping[str, Any]] = None

    @property
    def base(self) -> str:
        return str(self.__request.base_url)

    @property
    def url_path(self) -> str:
        scope = self.__request.scope
        root_path = scope.get("root_path")
        url = self.base
        if root_path:
            url = urljoin(url, root_path)
        return url

    @property
    def host(self) -> str:
        return self.__request.base_url.netloc

    @property
    def environ(self) -> MutableMapping[str, Any]:
        """
        Fallback WSGI environ.

        This is not a full environ, there is no body. This is only meant to make routes.url_for work.
        """
        if self.__environ is None:
            self.__environ = build_environ(self.__request.scope, None)  # type: ignore[arg-type]
        return self.__environ


class GalaxyASGIResponse(GalaxyAbstractResponse):
    """Wrapper around Starlette/FastAPI Response object.

    Implements the GalaxyAbstractResponse interface to provide access to some properties
    of the response object commonly used."""

    def __init__(self, response: Response):
        self.__response = response

    @property
    def headers(self):
        return self.__response.headers


DependsOnUser = cast(Optional[User], Depends(get_user))


def get_current_history_from_session(galaxy_session: Optional[model.GalaxySession]) -> Optional[model.History]:
    if galaxy_session:
        return galaxy_session.current_history
    return None


def get_trans(
    request: Request,
    response: Response,
    app: StructuredApp = DependsOnApp,
    user=cast(Optional[User], Depends(get_user)),
    galaxy_session=cast(Optional[model.GalaxySession], Depends(get_session)),
) -> SessionRequestContext:
    url_builder = UrlBuilder(request)
    galaxy_request = GalaxyASGIRequest(request)
    galaxy_response = GalaxyASGIResponse(response)
    return SessionRequestContext(
        app=app,
        user=user,
        galaxy_session=galaxy_session,
        url_builder=url_builder,
        request=galaxy_request,
        response=galaxy_response,
        history=get_current_history_from_session(galaxy_session),
    )


DependsOnTrans: SessionRequestContext = cast(SessionRequestContext, Depends(get_trans))


def get_admin_user(trans: SessionRequestContext = DependsOnTrans):
    if not trans.user_is_admin:
        raise AdminRequiredException(require_admin_message(trans.app.config, trans.user))
    return trans.user


AdminUserRequired = Depends(get_admin_user)


class BaseGalaxyAPIController(BaseAPIController):
    def __init__(self, app: StructuredApp):
        super().__init__(app)


class RestVerb(str, Enum):
    get = "GET"
    head = "HEAD"
    post = "POST"
    put = "PUT"
    patch = "PATCH"
    delete = "DELETE"
    options = "OPTIONS"


class Router(InferringRouter):
    """A FastAPI Inferring Router tailored to Galaxy."""

    def wrap_with_alias(self, verb: RestVerb, *args, alias: Optional[str] = None, **kwd):
        """
        Wraps FastAPI methods with additional alias keyword and require_admin handling.

        @router.get("/api/thing", alias="/api/deprecated_thing") will then create
        routes for /api/thing and /api/deprecated_thing.
        """
        kwd = self._handle_galaxy_kwd(kwd)
        include_in_schema = kwd.pop("include_in_schema", True)

        def decorate_route(route, include_in_schema=include_in_schema):
            # Decorator solely exists to allow passing `route_class_override` to add_api_route
            def decorated_route(func):
                self.add_api_route(
                    route,
                    endpoint=func,
                    methods=[verb],
                    include_in_schema=include_in_schema,
                    **kwd,
                )
                return func

            return decorated_route

        routes = []
        for path in self.construct_aliases(args[0], alias):
            if path != "/" and path.endswith("/"):
                routes.append(decorate_route(path, include_in_schema=False))
            else:
                routes.append(decorate_route(path))

        def dec(f):
            for route in routes:
                f = route(f)
            return f

        return dec

    @staticmethod
    def construct_aliases(path: str, alias: Optional[str]):
        yield path
        if path != "/" and not path.endswith("/"):
            yield f"{path}/"
        if alias:
            yield alias
            if not alias == "/" and not alias.endswith("/"):
                yield f"{alias}/"

    def get(self, *args, **kwd):
        """Extend FastAPI.get to accept a require_admin Galaxy flag."""
        return self.wrap_with_alias(RestVerb.get, *args, **kwd)

    def patch(self, *args, **kwd):
        """Extend FastAPI.patch to accept a require_admin Galaxy flag."""
        return self.wrap_with_alias(RestVerb.patch, *args, **kwd)

    def put(self, *args, **kwd):
        """Extend FastAPI.put to accept a require_admin Galaxy flag."""
        return self.wrap_with_alias(RestVerb.put, *args, **kwd)

    def post(self, *args, **kwd):
        """Extend FastAPI.post to accept a require_admin Galaxy flag."""
        return self.wrap_with_alias(RestVerb.post, *args, **kwd)

    def delete(self, *args, **kwd):
        """Extend FastAPI.delete to accept a require_admin Galaxy flag."""
        return self.wrap_with_alias(RestVerb.delete, *args, **kwd)

    def options(self, *args, **kwd):
        return self.wrap_with_alias(RestVerb.options, *args, **kwd)

    def head(self, *args, **kwd):
        return self.wrap_with_alias(RestVerb.head, *args, **kwd)

    def _handle_galaxy_kwd(self, kwd):
        require_admin = kwd.pop("require_admin", False)
        if require_admin:
            if "dependencies" in kwd:
                kwd["dependencies"].append(AdminUserRequired)
            else:
                kwd["dependencies"] = [AdminUserRequired]

        return kwd

    @property
    def cbv(self):
        """Short-hand for frequently used Galaxy-pattern of FastAPI class based views.

        Creates a class-based view for for this router, for more information see:
        https://fastapi-utils.davidmontague.xyz/user-guide/class-based-views/
        """
        return cbv(self)


class APIContentTypeRoute(APIRoute):
    """
    Determines endpoint to match using content-type.
    """

    match_content_type: str

    def accept_matches(self, scope: Scope) -> Tuple[Match, Scope]:
        content_type_header = Headers(scope=scope).get("content-type", None)
        if not content_type_header:
            return Match.PARTIAL, scope
        if self.match_content_type not in content_type_header:
            return Match.NONE, scope
        return Match.FULL, scope

    def matches(self, scope: Scope) -> Tuple[Match, Scope]:
        accept_match, accept_scope = self.accept_matches(scope)
        if accept_match == Match.NONE:
            return accept_match, accept_scope
        match, child_scope = super().matches(accept_scope)
        return (
            match if match.value < accept_match.value else accept_match,
            child_scope,
        )


def as_form(cls: Type[BaseModel]):
    """
    Adds an as_form class method to decorated models. The as_form class method
    can be used with FastAPI endpoints.

    See https://github.com/tiangolo/fastapi/issues/2387#issuecomment-731662551
    """
    new_params = [
        inspect.Parameter(
            field.alias,
            inspect.Parameter.POSITIONAL_ONLY,
            default=(Form(field.default) if not field.required else Form(...)),
        )
        for field in cls.__fields__.values()
    ]

    async def _as_form(**data):
        try:
            return cls(**data)
        except ValidationError as e:
            raise RequestValidationError(e.raw_errors)

    sig = inspect.signature(_as_form)
    sig = sig.replace(parameters=new_params)
    _as_form.__signature__ = sig  # type: ignore[attr-defined]
    cls.as_form = _as_form  # type: ignore[attr-defined]
    return cls


async def try_get_request_body_as_json(request: Request) -> Optional[Any]:
    """Returns the request body as a JSON object if the content type is JSON."""
    if "application/json" in request.headers.get("content-type", ""):
        body = await request.json()
        return body
    return None


search_description_template = Template(
    """A mix of free text and GitHub-style tags used to filter the index operation.

## Query Structure

GitHub-style filter tags (not be confused with Galaxy tags) are tags of the form
`<tag_name>:<text_no_spaces>` or `<tag_name>:'<text with potential spaces>'`. The tag name
*generally* (but not exclusively) corresponds to the name of an attribute on the model
being indexed (i.e. a column in the database).

If the tag is quoted, the attribute will be filtered exactly. If the tag is unquoted,
generally a partial match will be used to filter the query (i.e. in terms of the implementation
this means the database operation `ILIKE` will typically be used).

Once the tagged filters are extracted from the search query, the remaining text is just
used to search various documented attributes of the object.

## GitHub-style Tags Available

${tags}

## Free Text

Free text search terms will be searched against the following attributes of the
${model_name}s: ${freetext}.

"""
)


class IndexQueryTag(NamedTuple):
    tag: str
    description: str
    alias: Optional[str] = None
    admin_only: bool = False

    def as_markdown(self):
        desc = self.description
        alias = self.alias
        if alias:
            desc += f" (The tag `{alias}` can be used a short hand alias for this tag to filter on this attribute.)"
        if self.admin_only:
            desc += " This tag is only available for requests using admin keys and/or sessions."
        return f"`{self.tag}`\n: {desc}"


def search_query_param(model_name: str, tags: list, free_text_fields: list) -> Optional[str]:
    tags_markdown_str = "\n\n".join([t.as_markdown() for t in tags])
    description = search_description_template.safe_substitute(
        model_name=model_name, tags=tags_markdown_str, freetext=", ".join([f"`{t}`" for t in free_text_fields])
    )
    return Query(
        default=None,
        title="Search query.",
        description=description,
    )



"""
Plugins resource control over the API.
"""
import logging

from galaxy import exceptions
from galaxy.managers import (
    hdas,
    histories,
)
from galaxy.util import asbool
from galaxy.web import (
    expose_api,
    expose_api_anonymous_and_sessionless,
)
from . import (
    BaseGalaxyAPIController,
    depends,
)

log = logging.getLogger(__name__)


class PluginsController(BaseGalaxyAPIController):
    """
    RESTful controller for interactions with plugins.
    """

    hda_manager: hdas.HDAManager = depends(hdas.HDAManager)
    history_manager: histories.HistoryManager = depends(histories.HistoryManager)

    @expose_api_anonymous_and_sessionless
    def index(self, trans, **kwargs):
        """
        GET /api/plugins:
        """
        registry = self._get_registry()
        dataset_id = kwargs.get("dataset_id")
        if dataset_id is not None:
            hda = self.hda_manager.get_accessible(self.decode_id(dataset_id), trans.user)
            return registry.get_visualizations(trans, hda)
        else:
            embeddable = asbool(kwargs.get("embeddable"))
            return registry.get_plugins(embeddable=embeddable)

    @expose_api
    def show(self, trans, id, **kwargs):
        """
        GET /api/plugins/{id}:
        """
        registry = self._get_registry()
        history_id = kwargs.get("history_id")
        if history_id is not None:
            history = self.history_manager.get_owned(
                trans.security.decode_id(history_id), trans.user, current_history=trans.history
            )
            result = {"hdas": []}
            for hda in history.contents_iter(types=["dataset"], deleted=False, visible=True):
                if registry.get_visualization(trans, id, hda):
                    result["hdas"].append({"id": trans.security.encode_id(hda.id), "name": hda.name})
        else:
            result = registry.get_plugin(id).to_dict()
        return result

    def _get_registry(self):
        if not self.app.visualizations_registry:
            raise exceptions.MessageException("The visualization registry has not been configured.")
        return self.app.visualizations_registry
